use crate::models::deck::{Card, Deck};
use std::collections::HashMap;

#[derive(Debug)]
pub enum HandSort {
    None,
    Value,
    Rank,
}

#[derive(Debug)]
pub struct Hand {
    pub cards: Vec<Card>,
    pub card_sort: HandSort,
    pub hand_value: u8,
}

impl Hand {
    pub fn new() -> Hand {
        Hand {
            cards: Vec::new(),
            card_sort: HandSort::None,
            hand_value: 0,
        }
    }

    pub fn draw(&mut self, count: u8, deck: &mut Deck) {
        for _ in 0..count {
            let card = deck.cards.pop().unwrap();
            self.cards.push(card);
        }
    }

    pub fn sort(&mut self, hand_sort: HandSort) {
        match hand_sort {
            HandSort::Value => self.sort_by_value(),
            HandSort::Rank => self.sort_by_rank(),
            HandSort::None => (),
        }
    }

    pub fn sort_by_value(&mut self) {
        self.cards.sort_by(|a, b| a.value.cmp(&b.value));
        self.card_sort = HandSort::Value;
    }

    pub fn sort_by_rank(&mut self) {
        //self.cards.sort_by(|a, b| a.rank.cmp(&b.rank));
        self.card_sort = HandSort::Rank;
    }

    /// Checks all cards and returns total amount of pairs in the hand
    fn pair_count(&self) -> u8 {
        let mut pairs = 0;

        let mut value_counts: HashMap<u8, u8> = HashMap::new();
        for card in self.cards.iter() {
            value_counts.insert(card.value, value_counts.get(&card.value).unwrap_or(&0) + 1);
        }

        for i in value_counts.values() {
            if i > &1 {
                pairs = pairs + 1;
            }
        }

        pairs
    }

    /// Checks if the current hand has only one pair.  \
    /// One pair is a hand that contains two cards of one rank and three cards of three
    /// other ranks, such as `4♥ 4♠ K♠ 10♦ 5♠`
    pub fn has_single_pair(&mut self) -> bool {
        if !matches!(self.card_sort, HandSort::Value) {
            self.sort(HandSort::Value);
        }

        if self.pair_count() == 1 {
            return true;
        }

        false
    }

    /// Checks if the current hand has exactly two pairs. \
    /// Two pair is a hand that contains two cards of one rank, two cards of another rank,
    /// and one card of a third rank, such as `J♥ J♣ 4♣ 4♠ 9♥`
    pub fn has_two_pair(&mut self) -> bool {
        if !matches!(self.card_sort, HandSort::Value) {
            self.sort(HandSort::Value);
        }

        if self.pair_count() == 2 {
            return true;
        }

        false
    }

    /// Checks if the current hand has a three of a kind. \
    /// Three of a kind is a hand that contains three cards of one rank and
    /// two cards of two other ranks, such as `2♦ 2♠ 2♣ K♠ 6♥`
    pub fn has_three_of_a_kind(&mut self) -> bool {
        if !matches!(self.card_sort, HandSort::Value) {
            self.sort(HandSort::Value);
        }
        for i in 2..self.cards.len() {
            if self.cards[i].rank == self.cards[i - 1].rank &&
                self.cards[i].rank == self.cards[i - 2].rank {
                return true;
            }
        }

        false
    }

    /// Checks if the current hand has a straight. \
    /// A straight is a hand that contains five cards of sequential rank,
    /// not all the same suit, such as `7♣ 6♠ 5♠ 4♥ 3♥`
    pub fn has_straight(&mut self) -> bool {
        if !matches!(self.card_sort, HandSort::Value) {
            self.sort(HandSort::Value);
        }

        let hand_size = self.cards.len();

        // Handle Ace low straight
        if self.cards[0].value == 2 && self.cards[hand_size - 1].value == 14 {
            self.cards[hand_size - 1].value = 1;
            self.cards.rotate_right(1)
        }

        for i in 1..self.cards.len() {
            if self.cards[i].value != self.cards[i - 1].value + 1 {
                return false;
            }
        }

        true
    }

    /// Checks if the current hand has a flush. \
    /// A flush is a hand that contains five cards, all the same suit,
    /// and not all of sequential rank, such as `K♣ 10♣ 7♣ 6♣ 4♣`
    pub fn has_flush(&self) -> bool {
        for i in 1..self.cards.len() {
            if self.cards[i].suit != self.cards[i - 1].suit {
                return false;
            }
        }

        true
    }

    /// Checks if the current hand has a full house. \
    /// A full house is a hand that contains three cards of one
    /// rank and two cards of another rank, such as `3♣ 3♠ 3♦ 6♣ 6♥`
    pub fn has_full_house(&mut self) -> bool {
        if !matches!(self.card_sort, HandSort::Value) {
            self.sort(HandSort::Value);
        }

        let mut has_pair = false;
        let mut has_three_of_a_kind = false;

        // Check the first two positions for a pair to let the loop check for 2/3 of a kind
        if self.cards[0].rank == self.cards[1].rank { has_pair = true; }
        for i in 2..self.cards.len() {
            if self.cards[i].rank == self.cards[i - 1].rank {
                has_pair = true;
            }

            if self.cards[i].rank == self.cards[i - 1].rank &&
                self.cards[i].rank == self.cards[i - 2].rank {
                has_three_of_a_kind = true;
            }
        }
        has_pair && has_three_of_a_kind
    }

    /// Checks if the current hand has a four of a kind. \
    /// Four of a kind is a hand that contains four cards of one rank
    /// and one card of another rank, such as `9♣ 9♠ 9♦ 9♥ J♥`
    pub fn has_four_of_a_kind(&mut self) -> bool {
        if !matches!(self.card_sort, HandSort::Value) {
            self.sort(HandSort::Value);
        }

        for i in 3..self.cards.len() {
            if self.cards[i].rank == self.cards[i - 1].rank &&
                self.cards[i].rank == self.cards[i - 2].rank &&
                self.cards[i].rank == self.cards[i - 3].rank
            {
                return true;
            }
        }

        false
    }

    /// Checks if the current hand has a straight flush. \
    /// A straight flush is a hand that contains five cards of sequential rank,
    /// all the same suit, such as `Q♥ J♥ 10♥ 9♥ 8♥`
    pub fn has_straight_flush(&mut self) -> bool {
        if !matches!(self.card_sort, HandSort::Value) {
            self.sort(HandSort::Value);
        }

        let mut has_flush = true;
        let mut has_straight = true;

        let suit = &self.cards[0].suit;

        for i in 1..self.cards.len() {
            if &self.cards[i].suit != suit {
                has_flush = false;
                break;
            }
            if self.cards[i].value != self.cards[i - 1].value + 1 {
                has_straight = false;
                break;
            }
        }

        has_flush && has_straight
    }

    /// Checks if the current hand has a royal flush, the best possible hand. \
    /// A royal flush is a hand that contains an ace-high straight flush, such as `A♦ K♦ Q♦ J♦ 10♦`
    pub fn has_royal_flush(&mut self) -> bool {
        if !matches!(self.card_sort, HandSort::Value) {
            self.sort(HandSort::Value);
        }

        let mut has_flush = true;
        let mut has_straight = true;
        let suit = &self.cards[0].suit;

        for i in 1..self.cards.len() {
            if &self.cards[i].suit != suit {
                has_flush = false;
                break;
            }
            if self.cards[i].value != self.cards[i - 1].value + 1 {
                has_straight = false;
                break;
            }
        }

        has_flush && has_straight && self.cards[0].value == 10
    }
}
